from hashformers.beamsearch.minicons_lm import MiniconsLM

class BertLM(MiniconsLM):
    """
    Implements a BERT-based language model scorer, to compute sentence probabilities.
    This class uses a transformer-based Masked Language Model (MLM) for scoring. 
    
    Args:
        model_name_or_path (str): Identifier for the model to be loaded, which can be a model 
            name or the path to the directory where the model is stored.
        gpu_batch_size (int, optional): The size of the batch to be processed on the GPU. 
            Defaults to 1.
        gpu_id (int, optional): Identifier of the GPU device to be used. Defaults to 0.

    """
    def __init__(self, model_name_or_path, gpu_batch_size=1, gpu_id=0):
        super().__init__(
            model_name_or_path=model_name_or_path,
            device='cuda',
            gpu_batch_size=gpu_batch_size,
            model_type='MaskedLMScorer'
        )


from hashformers.beamsearch.gpt2_lm import GPT2LM
from hashformers.beamsearch.minicons_lm import MiniconsLM
from hashformers.beamsearch.bert_lm import BertLM

class ModelLM(object):
    """
    A Language Model (LM) class that supports both GPT2 and BERT models.

    This class acts as a wrapper around the GPT2LM and BertLM classes, providing
    a unified interface for interacting with either type of model. The specific
    type of model to use is determined by the 'model_type' argument provided during
    initialization.

    Args:
        model_name_or_path (str, optional): The name or path of the pre-trained model.
        model_type (str, optional): The type of the model to use. Should be either 'gpt2' or 'bert'.
        device (str, optional): The device on which to run the computations. Defaults to None which implies CPU.
        gpu_batch_size (int, optional): The batch size to use when performing computations on the GPU.
        gpu_id (int, optional): The ID of the GPU to use. Only relevant if 'model_type' is 'bert'. Default is 0.

    Raises:
        ValueError: If an unsupported 'model_type' is provided.
    """
    def __init__(self, model_name_or_path=None, model_type=None, device=None, gpu_batch_size=None, gpu_id=0):
        self.gpu_batch_size = gpu_batch_size
        if model_type is None:
            self.model = None
        elif model_type == 'gpt2':
            self.model = GPT2LM(model_name_or_path, device=device, gpu_batch_size=gpu_batch_size)
        elif model_type == 'bert':
            self.model = BertLM(model_name_or_path, gpu_batch_size=gpu_batch_size, gpu_id=gpu_id)
        elif model_type == 'seq2seq':
            self.model = MiniconsLM(model_name_or_path, device=device, gpu_batch_size=gpu_batch_size, model_type='Seq2SeqScorer')
        elif model_type == 'masked':
            self.model = MiniconsLM(model_name_or_path, device=device, gpu_batch_size=gpu_batch_size, model_type='MaskedLMScorer')
        elif model_type == 'incremental':
            self.model = MiniconsLM(model_name_or_path, device=device, gpu_batch_size=gpu_batch_size, model_type='IncrementalLMScorer')
        else:
            self.model = MiniconsLM(model_name_or_path, device=device, gpu_batch_size=gpu_batch_size, model_type=model_type)


from minicons import scorer
from torch.utils.data import DataLoader
import warnings
import math

class MiniconsLM(object):

    def __init__(self, model_name_or_path, device='cuda', gpu_batch_size=20, model_type='IncrementalLMScorer'):
        self.scorer = getattr(scorer, model_type)(model_name_or_path, device)
        self.gpu_batch_size = gpu_batch_size
        self.model_type = model_type
    
    def get_probs(self, list_of_candidates):
        probs = []
        dl = DataLoader(list_of_candidates, batch_size=self.gpu_batch_size)
        for batch in dl:
            probs.extend(self.get_batch_scores(batch))
        return probs
    
    def incremental_sequence_score(self, batch):
        tokens = self.scorer.prepare_text(batch, bos_token=True, eos_token=True)
        stats = self.scorer.compute_stats(tokens, prob=True)
        log_stats = [ [ math.log(x) for x in sequence ] for sequence in stats ]
        sum_log_stats = [ sum(x) for x in log_stats ]
        pos_sum_log_stats = [ 1 - x for x in sum_log_stats ]
        return pos_sum_log_stats

    def get_batch_scores(self, batch):
        if self.model_type == 'IncrementalLMScorer':
            return self.incremental_sequence_score(batch)
        elif self.model_type == 'MaskedLMScorer':
            return self.scorer.sequence_score(batch, reduction = lambda x: x.sum(0).item())
        elif self.model_type == 'Seq2SeqScorer':
            return self.scorer.sequence_score(batch, source_format = 'blank')
        else:
            warnings.warn(f"Model type {self.model_type} not implemented. Assuming reduction = lambda x: x.sum(0).item()")
            return self.scorer.sequence_score(batch, reduction = lambda x: x.sum(0).item())


from setuptools import find_packages, setup

setup(
   name='hashformers',
   version='2.0.0',
   author='Ruan Chaves Rodrigues',
   author_email='ruanchave93@gmail.com',
   description='Word segmentation with transformers',
   packages=find_packages('src'),
   package_dir={'': 'src'},
   install_requires=[
   "minicons",
   "twitter-text-python",
   "pandas"
   ]
)



from hashformers.beamsearch.data_structures import enforce_prob_dict
from hashformers.experiments.utils import build_ensemble_df

def run_ensemble(
    a_diff,
    b_diff,
    a_rank,
    b_rank,
    alpha=0.0,
    beta=0.0):
    """
    Computes the ensemble output using given differences and ranks with weights alpha and beta.

    Args:
        a_diff (array-like): Differences corresponding to 'a'.
        b_diff (array-like): Differences corresponding to 'b'.
        a_rank (array-like): Ranks corresponding to 'a'.
        b_rank (array-like): Ranks corresponding to 'b'.
        alpha (float, optional): The weight for 'a_diff'. Default is 0.0.
        beta (float, optional): The weight for 'b_diff'. Default is 0.0.

    Returns:
        array-like: An array-like object representing the ensemble output.
    """
    delta = alpha * a_diff - beta * b_diff
    decision = (delta < 0).astype(int)
    negation =  (~(delta < 0)).astype(int)
    output = a_rank * negation + b_rank * decision
    
    return output

def top2_ensemble(
    dict_1, 
    dict_2, 
    alpha=0.2, 
    beta=0.1):
    """
    Computes the ensemble of two given dictionaries using the specified weights alpha and beta.

    Args:
        dict_1 (dict): The first input dictionary.
        dict_2 (dict): The second input dictionary.
        alpha (float, optional): The weight for differences in 'dict_1'. Default is 0.2.
        beta (float, optional): The weight for differences in 'dict_2'. Default is 0.1.

    Returns:
        DataFrame: A pandas DataFrame representing the ensemble of the two input dictionaries.
    """
    a = enforce_prob_dict(dict_1).to_dataframe(
        characters_field="hashtag"
    )
    b = enforce_prob_dict(dict_2).to_dataframe(
        characters_field="hashtag"
    )

    ensemble_df = build_ensemble_df(a, b)

    ref_diff = ensemble_df["diff"].values
    aux_diff = ensemble_df["diff_2"].values
    ref_rank = ensemble_df["rank"].values
    aux_rank = ensemble_df["rank_2"].values

    ensemble_df["ensemble_rank"] = run_ensemble(
        ref_diff,
        aux_diff,
        ref_rank,
        aux_rank,
        alpha=alpha,
        beta=beta
    )

    return ensemble_df

class Top2_Ensembler(object):
    """
    A class that provides a method to run the ensemble of a segmenter run and a reranker run.

    Args:
        None
    """

    def __init__(self):
        pass

    def run(self, segmenter_run, reranker_run, alpha=0.222, beta=0.111):
        """
        Runs the ensemble of a segmenter run and a reranker run.

        Args:
            segmenter_run (dict or ProbabilityDictionary): The result of a segmenter run.
            reranker_run (dict or ProbabilityDictionary): The result of a reranker run.
            alpha (float, optional): The weight for differences in 'segmenter_run'. Default is 0.222.
            beta (float, optional): The weight for differences in 'reranker_run'. Default is 0.111.

        Returns:
            ProbabilityDictionary: A ProbabilityDictionary representing the ensemble of the segmenter and reranker runs.
        """
        ensemble = top2_ensemble(
            segmenter_run,
            reranker_run,
            alpha=alpha,
            beta=beta
        )

        ensemble_prob_dict = enforce_prob_dict(
            ensemble,
            score_field="ensemble_rank")
        
        return ensemble_prob_dict